#!/usr/bin/env python3
# ^^^^^^^^^^^^^^^^^^^^
# Need to define this since protoc expects the plugin to be an executable.

# https://expobrain.net/2015/09/13/create-a-plugin-for-google-protocol-buffer/
# https://golangexample.com/create-a-protocol-buffers-protobuf-plugin-which-is-executed-with-the-protoc-compile/
# https://rotemtam.com/2021/03/22/creating-a-protoc-plugin-to-gen-go-code/
# Protofile for CodeGeneratorRequest/Response https://github.com/protocolbuffers/protobuf/blob/master/src/google/protobuf/compiler/plugin.proto
# Description of SourceCodeInfo: https://github.com/protocolbuffers/protobuf/blob/master/src/google/protobuf/descriptor.proto

import os
import ast
import sys
import jinja2
from typing import Iterator, Tuple

file_x = lambda *x: os.path.join(os.path.dirname(__file__), *x)

from contextlib import contextmanager
from google.protobuf.compiler.plugin_pb2 import CodeGeneratorRequest, CodeGeneratorResponse

def debug(x):
  with open("./debug.txt", "w") as f:
    f.write(str(x))
      

class YQLNodeVisitor(ast.NodeVisitor):
  def __init__(self, functions_present: set, code_lines) -> None:
    self.imports = []
    self.fn_code ={}
    self.functions_present = functions_present
    self.code_lines = code_lines

  def _get_code_portion(self, lineno, col_offset, end_lineno, end_col_offset, **_):
    """Returns code piece for any section given all the lines it contains"""
    sl, so, el, eo = lineno, col_offset, end_lineno, end_col_offset
    if sl == el:
      return self.code_lines[sl-1][so:eo]
    code = ""
    for i in range(sl - 1, el, 1):
      if i == sl - 1:
        code += self.code_lines[i][so:]
      elif i == el - 1:
        code += "\n" + self.code_lines[i][:eo]
      else:
        code += "\n" + self.code_lines[i]
    return code

  def visit_FunctionDef(self, node: ast.FunctionDef):
    if node.name in self.functions_present:
      fn_code = self._get_code_portion(
        lineno = node.body[0].__dict__["lineno"],
        col_offset = node.body[0].__dict__["col_offset"],
        end_lineno = node.body[-1].__dict__["end_lineno"],
        end_col_offset = node.body[-1].__dict__["end_col_offset"],
      )
      if fn_code == "raise NotImplementedError":
        # there is nothing in this function so we skip this
        return
      fn_code = " " * node.body[0].__dict__["col_offset"] + fn_code
      self.fn_code[node.name] = fn_code


class Env:
  YQL_FOLDER: str = lambda x: os.getenv("YQL_FOLDER", x)
  YQL_PBREL: str = lambda x: os.getenv("YQL_PBREL", x)

@contextmanager
def code_generation() -> Iterator[Tuple[CodeGeneratorRequest, CodeGeneratorResponse]]:
    if len(sys.argv) > 1 and sys.argv[1] in ("-V", "--version"):
      print("yql-protobuf-compiler version 0.0.1")
      sys.exit(0)

    # create Protos
    request = CodeGeneratorRequest()
    response = CodeGeneratorResponse()
    response.supported_features |= (CodeGeneratorResponse.FEATURE_PROTO3_OPTIONAL)

    request.ParseFromString(sys.stdin.buffer.read())

    yield request, response

    # iterate over all the files in the data
    no_service_def = []
    multiple_service = []
    for proto_file in request.proto_file:

      if len(proto_file.service) == 0:
        no_service_def.append(proto_file)
        continue
      
      if len(proto_file.service) > 1:
        multiple_service.append(proto_file)
        response.error = "Multiple service definitions found, should contain only one"
        sys.stdout.buffer.write(response.SerializeToString())
        return
      sys.stdout.buffer.write(response.SerializeToString())
      
      # create the data points
      proto_imports = {}
      all_services = set()
      all_protos = set()
      
      # create all the data needed to write the file
      # TODO: @yashbonde: figure out how to get the comments as well for the parts
      trg_folder = Env.YQL_FOLDER("")
      pb_relative = Env.YQL_PBREL("")
      if pb_relative:
        pb_relative += "."
      proto_folder, proto_name = os.path.split(proto_file.name)
      trg_folder = trg_folder or proto_folder

      run_server = os.path.join(trg_folder, "server.py")
      trg_server = os.path.join(trg_folder, proto_name.split(".")[0] + "_server.py")
      trg_client = os.path.join(trg_folder, proto_name.split(".")[0] + "_client.py")

      service = proto_file.service[0]
      service_tuples = []
      for method in service.method:
        # process the input message information
        _in_var =  method.input_type.replace(".", "_")
        _in_proto = method.input_type.strip(".")
        _in_proto_message = _in_proto.split(".")[-1]
        proto_imports.setdefault(pb_relative + proto_name.split(".")[0] + "_pb2", []).append(_in_proto_message)

        # process the output message information
        _out_var = method.output_type.replace(".", "_")
        _out_proto = method.output_type.strip(".")
        _out_proto_message = _out_proto.split(".")[-1]
        proto_imports.setdefault(pb_relative + proto_name.split(".")[0] + "_pb2", []).append(_out_proto_message)

        # create the service tuple
        service_tuples.append((method.name, _in_var, _in_proto_message, _out_var, _out_proto_message))
        all_services.add(method.name)
        all_protos.add(_in_proto_message)
        all_protos.add(_out_proto_message)

      # now create all the items needed by our Jinja Renderer
      dependency_strings = [f"from {'.'.join(x.split('/'))[:-6]}_pb2 import *" for x in proto_file.dependency]
      imports_strings = [f"from {k} import *" for k,v in proto_imports.items()]
      all_imports = dependency_strings + imports_strings
      all_services = sorted(all_services)
      all_protos = sorted(all_protos)

      if os.path.exists(trg_server):
        # the server file already exists, we should update that instead of overwriting it, since most of time
        # the update is going to be a small update like addition of a new service
        os.system(f"mv {trg_server} {trg_server}.back")
        with open(f"{trg_server}.back", "r") as f:
          code = f.read()
          node = ast.parse(code)
          code_lines = code.splitlines()

        nv = YQLNodeVisitor(functions_present=set([x[0] for x in service_tuples]), code_lines=code_lines)
        nv.visit(node)

        # for now we are not doing anything, waiting on some more ideas for this

      # load the jinja templates
      with open(file_x("assets", "server_stub.jinja"), "r") as src, open(trg_server, "w") as trg:
        trg.write(jinja2.Template(src.read()).render(
          service_tuples=service_tuples,
          all_imports=all_imports,
          service_name=service.name,
        ))

      with open(file_x("assets", "client_stub.jinja"), "r") as src, open(trg_client, "w") as trg:
        trg.write(jinja2.Template(src.read()).render(
          service_tuples=service_tuples,
          all_imports=all_imports,
          service_name = service.name,
          zip=zip
        ))

      with open(file_x("assets", "server.jinja"), "r") as src, open(run_server, "w") as trg:
        trg.write(jinja2.Template(src.read()).render(
          server_stub = proto_name.split(".")[0] + "_server",
          service_name = service.name,
          all_services=list(all_services),
          all_protos=list(all_protos),
          all_imports=all_imports,
        ))

      # embed yql in target folder
      files_to_move = [file_x("__init__.py"), file_x("common.py"), file_x("rest_pb2.py"), file_x("rest_pb2.pyi")]
      if not os.path.exists(f"{trg_folder}/yql"):
        os.system(f"mkdir {trg_folder}/yql")
        os.system(f"cp {' '.join(files_to_move)} {trg_folder}/yql")


def main():
  with code_generation() as (request, response):
    pass

if __name__ == "__main__":
  main()



import os

from json import dumps
from typing import Dict, Any, List, Optional, Union
from nbox.utils import logger, env as nbx_env
from nbox.utils import get_files_in_folder
from nbox import Instance
from nbox.auth import secret
from nbox.nbxlib.tracer import Tracer
from requests import Session

import dainik.utils as U
from dainik.utils import env
from dainik.lmao_client import *
from dainik.relic import RelicStore, get_relic_file
from dainik.proto.lmao_pb2 import AgentDetails
from dainik.git_utils import get_git_details
from dainik.metrics.system import SystemMetricsLogger

DEBUG_LOG_EVERY = 5
INFO_LOG_EVERY = 100

class Dainik():
  def __init__(self, workspace_id: str, *, local: bool = False) -> None:
    self.workspace_id = workspace_id
    self.local = local

    self._initialized = False
    self.nbx_job_folder = nbx_env.NBOX_JOB_FOLDER("")
    self._total_logged_elements = 0 # this variable keeps track for logging
    self.completed = False
    self.tracer = None
    self.username = secret.get("username")
    self.relic = None

    if not local:
      self._create_connection(workspace_id)
    else:
      self.lmao = LMAO_Stub(url = "http://127.0.0.1:8080", session = Session())

  def _create_connection(self, workspace_id: str):
    # prepare the URL
    # id_or_name, workspace_id = U.split_iw(instance_id)
    id_or_name = f"monitoring-{workspace_id}"
    logger.info(f"id_or_name: {id_or_name}")
    logger.info(f"workspace_id: {workspace_id}")
    instance = Instance(id_or_name, workspace_id)
    try:
      open_data = instance.open_data
    except AttributeError:
      raise Exception(f"Is instance '{instance.project_id}' running?")
    build = "build"
    if "app.c." in secret.get("nbx_url"):
      build = "build.c"
    url = f"https://server-{open_data['url']}.{build}.nimblebox.ai/"
    logger.info(f"URL: {url}")

    # create a tracer object that will load all the information
    self.tracer = Tracer(start_heartbeat=False)
    self.username = self.tracer.job_proto.auth_info.username if self.tracer else self.username

    # create a session with the auth header
    _session = Session()
    _session.headers.update({
      "NBX-TOKEN": open_data["token"],
      "X-NBX-USERNAME": self.username,
    })

    # define the stub
    self.lmao = LMAO_Stub(url = url, session = _session)

  def init(self, project_name: Optional[str] = "", config: Dict[str, Any] = {}, project_id: Optional[str] = ""):
    if project_id:
      project_name = ""
      if project_name:
        logger.info("ignoring `project_name` as `project_id` was given")
    else:
      project_id = ""
      if not project_name:
        raise Exception("Provide one from `project_id` or `project_name`")

    # this is the config value that is used to store data on the plaform, user cannot be allowed to have
    # like a full access to config values
    log_config: Dict[str, Any] = {
      "user_config": config
    }

    # check if the current folder from where this code is being executed has a .git folder
    # NOTE: in case of NBX-Jobs the current folder ("./") is expected to contain git by default
    log_config["git"] = None
    if os.path.exists(".git"):
      log_config["git"] = get_git_details("./")

    # continue as before
    self.project_name = project_name
    self.project_id = project_id
    self.config = config
    self._agent_details = AgentDetails(
      nbx_job_id = self.tracer.job_id if self.tracer else "jj_guvernr",
      nbx_run_id = self.tracer.token if self.tracer else U.get_random_id(),
    )
    run_details = self.lmao.init_run(
      _InitRunRequest = InitRunRequest(
        agent_details=self._agent_details,
        created_at = U.get_timestamp(),
        project_name = self.project_name,
        project_id = self.project_id,
        config = dumps(log_config),
      )
    )

    if not run_details:
      # TODO: Make a custom exception of this
      raise Exception("Server Side exception has occurred, Check the log for details")

    # TODO: change f-string here when refactoring `run_id` to `exp_id`
    logger.info(f"Assigned run_id: {run_details.run_id}")
    self.run = run_details
    self._initialized = True

    # system metrics monitoring, by default is enabled optionally turn it off
    if not env.DK_DISABLE_SYSTEM_METRICS(False):
      SystemMetricsLogger(self)

    # now initialize the relic
    if not env.DK_DISABLE_RELICS(False):
      self.relic = RelicStore(self.workspace_id, relic_name = self.run.run_id, create = True)

  def log(self, y: Dict[str, Union[int, float, str]], step = None, *, log_type: str = RunLog.LogType.USER):
    if not self._initialized:
      raise Exception("Run not initialized, call init() first")
    if self.completed:
      raise Exception("Run already completed, cannot log more data!")

    if self._total_logged_elements % DEBUG_LOG_EVERY == 0:
      logger.debug(f"Logging: {y.keys()} | {self._total_logged_elements}")

    if self._total_logged_elements % INFO_LOG_EVERY == 0:
      logger.info(f"Logging: {y.keys()} | {self._total_logged_elements}")

    step = step if step is not None else U.get_timestamp()
    if step < 0:
      raise Exception("Step must be <= 0")
    run_log = RunLog(run_id = self.run.run_id, log_type=log_type)
    for k,v in y.items():
      record = U.get_record(k, v)
      record.step = step
      run_log.data.append(record)

    ack = self.lmao.on_log(_RunLog = run_log)
    if not ack.success:
      logger.error("  >> Server Error")
      for l in ack.message.splitlines():
        logger.error("  " + l)
      raise Exception("Server Error")

    self._total_logged_elements += 1

  def save_file(self, *files):
    """
    Register a file save and upload it by talking to the NimbleBox Relics Backend.
    `dk.save_file("foo.t", "/bar/", "baz.t", "/bar/roo/")`
    """
    logger.info(f"Saving files: {files}")
    if not self._initialized:
      raise Exception("Run not initialized, call init() first")

    # manage all the complexity of getting the list of RelicFile
    all_relic_files = []
    for folder_or_file in files:
      if os.path.isfile(folder_or_file):
        all_files = [folder_or_file]
      elif os.path.isdir(folder_or_file):
        all_files = get_files_in_folder(folder_or_file, self.username, self.workspace_id)
      else:
        raise Exception(f"File or Folder not found: {folder_or_file}")

      out = []
      for f in all_files:
        relic_file = get_relic_file(f, self.username, self.workspace_id)
        relic_file.relic_name = self.run.run_id
        out.append(relic_file)
      all_relic_files.extend((f, x) for f,x in zip(all_files, out))

    # define API function
    def _register_upload(file_list):
      ack = self.lmao.on_save(_FileList = file_list)
      if not ack.success:
        logger.error("  >> Server Error")
        for l in ack.message.splitlines():
          logger.error("  " + l)
        raise Exception("Server Error")
    
    if self.relic:
      for local, relic_file in all_relic_files:
        # upload and then register
        self.relic.upload_relic_file(local, relic_file)
        file_list = FileList(run_id=self.run.run_id)
        file_list.files.append(File(relic_file = relic_file))
        _register_upload(file_list)
    else:
      all_relics = all_relic_files
      file_list = FileList(run_id=self.run.run_id)
      for local, relic_file in all_relics:
        file_list.files.append(File(relic_file = relic_file))
        self.relic.upload_relic_file(local, relic_file)
      _register_upload(file_list)


  def end(self):
    logger.info("Ending run")
    ack = self.lmao.on_train_end(_Run = Run(run_id=self.run.run_id,))
    if not ack.success:
      logger.error("  >> Server Error")
      for l in ack.message.splitlines():
        logger.error("  " + l)
      raise Exception("Server Error")
    self.completed = True



class SearchSpaceExhausted(Exception):
  pass

class BaseSampler(object):
  """This is the base class for all the samplers"""
  def __init__(self) -> None:
    super().__init__()

  def build_search_space(self) -> None:
    """Subclass and overwrite your logic to build the search space as per your definitions. Default is No-Op"""
    pass
    

  def sample(self) -> dict:
    """Subclass and overwrite your logic to get one sample of data, this should return a
    dictionary of parameters to be used for the experiment"""
    raise NotImplementedError("Subclass must implement abstract method")

  def update(self) -> None:
    """Subclass and overwrite your logic to update any kind of internal state of the sampler, ex.
    grid search can double check with the NBX-Servers if everything is fine or not. Gets called
    after every sample returned. By default is a No-Op"""
    pass

  def __iter__(self) -> None:
    try:
      while True:
          yield self.sample()
    except SearchSpaceExhausted:
      return



import numpy as np
from typing import Dict, List, Union
from pydantic import BaseModel

from dainik.hyperopt.base import BaseSampler

class RandomParameter(BaseModel):
  """Create a parameters to be sampled. Provide either ``x`` or ``a`` ` and ``b``.

  Args:
    x (List[Union[int, float, str]]): This is the possible list of values for this parameter.
    a (Union[float, int]): This is the lower bound of the random distribution.
    b (Union[float, int]): This is the upper bound of the random distribution.
    dist (str, optional): This is the distribution to use. Defaults to "uniform".
  """
  x: List[Union[int, float, str]] = []
  a: Union[float, int] = None
  b: Union[float, int] = None
  dist: str = "uniform"


class RandomSampler(BaseSampler):
  applicable_distributions = [
    "uniform", "normal", "beta", "poisson",
    # "lognormal",  "gamma", "weibull",
  ]

  def __init__(self, params: Dict[str, RandomParameter], n_runs: int = 10, seed: int = 4) -> None:
    super().__init__()

    # sanity checks
    param_to_chose = set()
    for k, param in params.items():
      assert param.dist in self.applicable_distributions, f"Distribution {param.dist} is not supported"
      if param.x:
        assert param.a is None and param.b is None, "Cannot specify x and a/b"
        param_to_chose.add(k)
      else:
        assert param.a is not None and param.b is not None, "Either specify x or must specify a/b"
        assert param.a < param.b, "a must be smaller than b"
        param.a = max(0.1, param.a)
        assert param.a > 0, "a must be positive"
        assert param.b > 0, "b must be positive"

    self.params = params
    self.n_runs = n_runs
    self.seed = seed
    self.rng = np.random.RandomState(seed)
    self.param_to_chose = param_to_chose

  def sample(self) -> dict:
    ret_params = {}
    for k, param in self.params.items():
      if k in self.param_to_chose:
        # this is based on a predetermined list of events
        if param.dist == "uniform":
          # this is the default so just sample, add and continue
          v = self.rng.choice(param.x)
          ret_params[k] = v
          continue
        
        elif param.dist == "normal":
          p = self.rng.normal(0, 1, size = len(param.x))
        elif param.dist == "beta":
          p = self.rng.beta(0, 1, size = len(param.x))
        elif param.dist == "poisson":
          p = self.rng.poisson(size = len(param.x))

        # force normalise distribution
        p = p.astype(np.float32)
        if p.min() < 0:
          p -= p.min()
        p /= p.sum()
        v = self.rng.choice(a = param.x, p = p)
      else:
        # this is based on a random distribution
        if param.dist == "uniform":
          v = self.rng.uniform(param.a, param.b)
        elif param.dist == "normal":
          v = self.rng.normal(param.a, param.b)
        elif param.dist == "beta":
          v = self.rng.beta(param.a, param.b)
        elif param.dist == "poisson":
          v = self.rng.poisson(param.a)
      ret_params[k] = v
    return ret_params

  def __iter__(self):
    for _ in range(self.n_runs):
      yield self.sample()



import re
import os
from datetime import datetime, timezone
from uuid import uuid4

from typing import Dict, Any, Union
from dainik.proto.lmao_pb2 import Record

class env:
  DK_INSTANCE_NAME = lambda x: os.environ.get("DK_INSTANCE_NAME", x)
  DK_DISABLE_SYSTEM_METRICS = lambda x: os.environ.get("DK_DISABLE_SYSTEM_METRICS", x)
  DK_DISABLE_RELICS = lambda x: os.environ.get("DK_DISABLE_RELICS", x)

def get_random_id(n=10):
  return str(uuid4()).replace("-", "")[:n]

def get_timestamp():
  return int(datetime.now(timezone.utc).timestamp())

def get_record(k: str, v: Union[int, float, str]) -> Record:
  _tv = type(v)
  assert _tv in [int, float, str], f"[key = {k}] '{_tv}' is not a valid type"
  _vt = {
    int: Record.DataType.INTEGER,
    float: Record.DataType.FLOAT,
    str: Record.DataType.STRING,
  }[_tv]
  record = Record(key = k, value_type = _vt)
  if _tv == int:
    record.integer_data.append(v)
  elif _tv == float:
    record.float_data.append(v)
  elif _tv == str:
    record.string_data.append(v)
  return record

def split_iw(x: str):
  pat = r"([\w\-\s]+)@(\w{8})"
  out = re.findall(pat, x)
  if not out:
    raise ValueError(f"Incorrect instance_id {x}")
  id_or_name, workspace_id = out[0]
  return id_or_name, workspace_id



from itertools import product
from typing import Union
from pydantic import BaseModel

from nbox import logger
from dainik.hyperopt.base import BaseSampler, SearchSpaceExhausted
from dainik.hyperopt.opt_types import _v0_data, List, Tuple, Dict

class GridParameter(BaseModel):
  data: Union[_v0_data, List[_v0_data], Tuple[_v0_data]]


class GridSampler(BaseSampler):
  def __init__(self, params: Dict[str, GridParameter]) -> None:
    super().__init__()
    self.params = params

    # do sanity checks on the values like they are all iterable or not
    keys = []
    values = []
    total_possible_combinations = 1
    for k,v in self.params.items():
      v = v.data
      if isinstance(v, (int, float, str)):
        # this usually means some constant value
        keys.append(k)
        values.append([v])
        continue
      keys.append(k)
      values.append(v)
      total_possible_combinations *= len(v)
    
    if total_possible_combinations > 100:
      logger.warning(f"Total possible combinations: {total_possible_combinations}")
      logger.warning(f"This is above the guard rail of 100 combinations.")

    # now we can build the search space, which is an array of all the possible states
    # this key_indexes contains to which array is current sample referring to
    self._keys = keys
    self._values = product(*values)

  def sample(self) -> dict:
    ret_params = {}
    try:
      values = next(self._values)
    except StopIteration:
      raise SearchSpaceExhausted("Search space exhausted")
    for k,v in zip(self._keys, values):
      ret_params[k] = v
    return ret_params




# The idea is that you can add this is a middleware right into your stack and take full
# advantage of the NimbleBox LMAO system.

try:
  # importing like this is a bit of a hack, but it works since we don't need to add
  # starlette as a dependency
  from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
  from starlette.types import ASGIApp
except ImportError:
  raise ImportError("Please install starlette to use this middleware. pip install starlette")

class LMAOMiddleware(BaseHTTPMiddleware):
  def __init__(self, app: ASGIApp) -> None:
    """LMAO Middleware for Starlette plugs into the ASGI stack and sends metrics to NimbleBox LMAO.
    Such as request count, response count, request processing time, and exceptions raised."""
    # simply added here so in future we can add more options
    super().__init__(app)


from starlette_prometheus import metrics, PrometheusMiddleware

app.add_middleware(PrometheusMiddleware)
app.add_route("/metrics/", metrics)


# Parts of the code are taken from SpellML library
# I will drop a more details during release
# License Unkown

import re
from git import Repo
from nbox import logger

def get_git_details(folder):
  repo = Repo(folder)

  # check for any unstaged files
  uncommited_files = {}
  diff = repo.index.diff(None)
  for f in diff:
    path = f.a_path or f.b_path # when new file is added, a_path is None
    uncommited_files[path] = f.change_type
  if uncommited_files:
    logger.warning(f"Uncommited files: {uncommited_files}")

  # get the remote url
  try:
    remote_url = repo.remote().url

    # clean the remote_url because it can sometimes contain the repo token as well.
    # this can become security hazard. so if you have an example, that is not suppported:
    #   go ahead, make a PR!
    if "github.com" in remote_url:
      remote_url = re.sub(r"ghp_\w+@", "", remote_url)
  except ValueError:
    remote_url = None

  # get the size of the repository
  size = None
  for line in repo.git.count_objects("-v").splitlines():
    if line.startswith("size:"):
      size = int(line[len("size:") :].strip())
  if size > (1 << 30):
    logger.warning(f"Repository size over 1GB, you might want to work on it")

  return {
    "remote_url": remote_url,
    "branch": repo.active_branch.name,
    "commit": repo.head.commit.hexsha,
    "uncommited_files": uncommited_files,
    "untracked_files": repo.untracked_files,
    "size": size,
  }




import os
import requests

from nbox.auth import secret
from nbox.utils import get_files_in_folder
from nbox import nbox_ws_v1, logger
from copy import deepcopy

from dainik.relics_rpc_client import RelicStore_Stub
from dainik.proto.relics_rpc_pb2 import *
from dainik.proto.relics_pb2 import *


def get_relic_file(fpath: str, username: str, workspace_id: str):
  assert os.path.exists(fpath), f"File {fpath} does not exist"
  assert os.path.isfile(fpath), f"File {fpath} is not a file"

  # clean up fpath, remove any trailing slashes
  # trim any . or / from prefix and suffix
  fpath_cleaned = fpath.strip("./")
  file_stat = os.stat(fpath)
  return RelicFile(
    name = fpath_cleaned,
    created_on = int(file_stat.st_mtime),    # int
    last_modified = int(file_stat.st_mtime), # int
    size = max(1, file_stat.st_size),        # bytes
    username = username,
    type = RelicFile.RelicType.FILE,
    workspace_id = workspace_id,
  )

class RelicStore():
  def __init__(self, workspace_id: str, relic_name: str, create: bool = False):
    self.workspace_id = workspace_id
    self.relic_name = relic_name
    self.username = secret.get("username") # if its in the job then this part will automatically be filled

    # relic_path = nbox_ws_v1.u(workspace_id).relics # v1/{workspace_id}/relics/
    # self.stub = RelicStore_Stub(relic_path._url, deepcopy(relic_path._session))    
    self.stub = RelicStore_Stub("http://0.0.0.0:8081", deepcopy(nbox_ws_v1._session))    
    
    _relic = self.stub.get_relic_details(
      Relic(workspace_id=workspace_id, name=relic_name,)
    )
    print("_______", _relic)
    if _relic is None and create:
      # this means that a new one will have to be created
      logger.info(f"Creating new relic {relic_name}")
      self.relic = self.stub.create_relic(
        CreateRelicRequest(workspace_id=workspace_id, name = relic_name,)
      )
    else:
      self.relic = _relic

    print(self.relic)
  
  def __repr__(self):
    return f"RelicStore({self.workspace_id}, {self.relic_name})"

  def upload_relic_file(self, local_path: str, relic_file: RelicFile):
    if not relic_file.relic_name:
      raise ValueError("relic_name not set in RelicFile")

    # ideally this is a lot like what happens in nbox
    logger.info(f"Uploading {local_path} to {relic_file.name}")
    out = self.stub.create_file(_RelicFile = relic_file,)
    
    # do not perform merge here because "url" might get stored in MongoDB
    # relic_file.MergeFrom(out)
    logger.info(f"URL: {out.url}")
    with open(local_path, "rb") as f:
      r = requests.put(out.url, data = f)
    logger.info(f"Upload status: {r.status_code}")

  def download_relic_file(self, local_path: str, relic_file: RelicFile):
    if self.relic is None:
      raise ValueError("Relic does not exist! pass create=True")

    # ideally this is a lot like what happens in nbox
    logger.info(f"Downloading {local_path} from S3 ...")
    out = self.stub.download_file(_RelicFile = relic_file,)
    
    # do not perform merge here because "url" might get stored in MongoDB
    # relic_file.MergeFrom(out)
    logger.info(f"URL: {out.url}")
    with requests.get(out.url, stream=True) as r:
      r.raise_for_status()
      total_size = 0
      with open(local_path, 'wb') as f:
        for chunk in r.iter_content(chunk_size=8192): 
          # If you have chunk encoded response uncomment if
          # and set chunk_size parameter to None.
          #if chunk: 
          f.write(chunk)
          total_size += len(chunk)
    logger.info("Download status: OK")
    logger.info(f"    filepath: {local_path}")
    logger.info(f"  total_size: {total_size//1024} KB")

  def _put(self, local_path: str):
    # if self.relic is None:
    #   raise ValueError("Relic does not exist! pass create=True")

    # get the file
    relic_file = get_relic_file(local_path, self.username, self.workspace_id)
    relic_file.relic_name = self.relic_name
    self.upload_relic_file(local_path, relic_file)
  
  def _get(self, local_path: str):
    if self.relic is None:
      raise ValueError("Relic does not exist! pass create=True")

    # get the file
    relic_file = get_relic_file(local_path, self.username, self.workspace_id)
    relic_file.relic_name = self.relic_name
    self.download_relic_file(local_path, relic_file)



from dainik.hyperopt.base import BaseSampler, SearchSpaceExhausted
from dainik.hyperopt.random_sampler import RandomParameter, RandomSampler
from dainik.hyperopt.grid_sampler import GridSampler, GridParameter